/* Copyright (C) 2008, Cyril Hu (cyrilhu@gmail.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/* ADT of singly linked list using external storage model.
 * see http://en.wikipedia.org/wiki/Linked_list for detail.
 */
#if defined(ADT_DEBUG) || defined(ADT_SLLIST_DEBUG)
#include <stdio.h>
#endif
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "sllist.h"

/* Init SLL structure to zero. */
void sll_init(sll_t *sll)
{
	sll->head = sll->tail = NULL;
	sll->linklen = 0;
}

/* Duplicate another SLL, pointing to the same external storage. */
bool sll_dup(sll_t *dst, const sll_t *sll)
{
	sll_node_t *p = sll->head, *q, *t;
#if defined(ADT_DEBUG)
	unsigned i = 1;
#endif
	
	/* sanity check. */
	if(!p) {
		return false;
	}

	/* allocate dst head node. */
	t = dst->head = malloc(sizeof(sll_node_t));
	if(!t) {
		return false;
	}
	t->data = p->data, p = p->next;

	/* copy src nodes to dst SLL one by one. */
	while(p) {
		q = malloc(sizeof(sll_node_t));
		if(!q) {
			t->next = NULL;
			sll_destroy(dst);
			return false;
		}
		q->data = p->data, p = p->next;
		t->next = q, t = q;
	}

	/* assign dst tail node and linklen. */
	dst->tail = t, t->next = NULL;
	dst->linklen = sll->linklen;
	return true;
}

/* Put SLL content in reversed order. */
void sll_rev(sll_t *r)
{
	sll_node_t *t, *p = r->head, *q;

	/* sanity check. */
	if(!p) {
		return;
	}

	/* reverse sll->next position. */
	for(q = p->next; q; p = q, q = t) {
		t = q->next, q->next = p;
	}

	/* reposition head & tail. */
	r->tail = r->head, r->tail->next = NULL;
	r->head = p;
}

/* Delete SLL's current head node. */
void sll_delhead(sll_t *sll)
{
	sll_node_t *t = sll->head;
	if(t) {
		/* SLL contains only one node. */
		if(!--sll->linklen) {
			sll->tail = NULL;
		}
		sll->head = t->next;
		free(t);
	}
}

/* Delete SLL's current tail node. */
void sll_deltail(sll_t *sll)
{
	size_t len = sll->linklen;
	sll_node_t *t;

	/* sanity check. */
	if(!len) {
		return;
	}

	/* SLL contains only one node. */
	if(!--sll->linklen) {
		free(sll->head);
		sll->head = sll->tail = NULL;
		return;
	}

	/* position tail node. */
	for(t = sll->head; len-- > 2; t = t->next)
		;

	/* delete tail node. */
	free(t->next);
	sll->tail = t, t->next = NULL;
}
	
/* Add a new head node. */
bool sll_addhead(sll_t *sll, const void *data)
{
	sll_node_t *t = sll->head, *r;

	/* allocate new head node. */
	r = malloc(sizeof(sll_node_t));
	if(!r) {
		return false;
	}
	sll->head = r, r->data = data, r->next = t;

	/* SLL is empty. */
	if(!(sll->linklen++)) {
		sll->tail = r;
	}
	return true;
}

/* Add a new tail node (append). */
bool sll_addtail(sll_t *sll, const void *data)
{
	/* allocate new tail node. */
	sll_node_t *r = malloc(sizeof(sll_node_t));
	if(!r) {
		return false;
	}
	r->data = data, r->next = NULL;

	/* assign new tail node. */	
	if(sll->linklen++) {
		sll->tail = sll->tail->next = r;
	} else { /* SLL is empty. */
		sll->head = sll->tail = r;
	}
	return true;
}

/* Conditionally (callback returns true) delete SLL node(s),
 * return number of deleted node(s). */
size_t sll_node_cdel(sll_t *sll, bool (*test)(const void *))
{
	size_t cnt = 0;
	sll_node_t *q = sll->head, *t;

	/* head node matches. */
	while(q && (*test)(q->data)) {
		sll->head = q->next;
		free(q);
		q = sll->head, ++cnt;
	}

	/* head node doesn't match. */
	while(q) {
		t = q->next;
		if(!t) { /* this is the tail node. */
			break;
		} else if( (*test)(&t->data)) { /* next node matches. */
			q->next = t->next;
			free(t), ++cnt;
		} else { /* next node doesn't match. */
			q = t;
		}
	}

	sll->tail = q; 
	sll->linklen -= cnt;
	return cnt;
}

/* Conditionally (callback returns true) update SLL node content,
 * return number of updated node(s). */
size_t sll_node_cupdt(sll_t *sll, bool (*test)(const void *), const void *data)
{
	size_t cnt = 0;
	sll_node_t *q = sll->head;
	while(q) {
		if( (*test)(q->data)) {
			q->data = data, ++cnt;
		}
		q = q->next;
	}
	return cnt;
}

/* Insert a node right before matching node (callback returns true),
 * return true on success, otherwise false. */
bool sll_node_bins(sll_t *sll, bool (*test)(const void *), const void *data)
{
	sll_node_t *t = sll->head, *r, *q;

	/* sanity check. */
	if(!t) {
		return false;
	}

	/* head node matches. */
	if( (*test)(t->data)) {
		sll_addhead(sll, data);
		return true;
	}

	/* position matching node. */
	for(q=t->next; q; t = t->next, q = t->next) {
		if( (*test)(q->data)) {
			r = t->next = malloc(sizeof(sll_node_t));
			if(!r) {
				return false;
			}
			r->data = data, r->next = q;
			return true;
		}
	}
	return false;
}

/* Insert a node right after matching node (callback returns true),
 * return true on success, otherwise false. */
bool sll_node_ains(sll_t *sll, bool (*test)(const void *), const void *data)
{
	sll_node_t *q = sll->head, *t, *r; 

	for(; q; q = q->next) {
		if( (*test)((t=q)->data)) {
			
			/* allocate new  node. */
			r = malloc(sizeof(sll_node_t));
			if(!r) {
				return false;
			}
			t->next = r, r->data = data;
			r->next = q->next, ++sll->linklen;	

			/* assign new  node. */
			if(!q->next) {
				sll->tail = r;
			}
			return true;
		}
	}
	return false;
}

/* Peek specific node's element content,
 * return the pointer on valid offset, otherwise NULL. */
const void *sll_node_peek(sll_t *sll, size_t offset)
{
	/* sanity check. */
	if(offset >= sll->linklen) {
		return NULL;
	} else {
		size_t i = 0;
		sll_node_t *r = sll->head;
		while(i++ < offset) {
			r = r->next;
		}
		return r->data;
	}
}

/* Number of node(s) SLL contains. */
size_t sll_getlen(sll_t *sll)
{
	return sll->linklen;
}

/* Sort SLL content by exchanging const pointers to external storage. */
void sll_sort(sll_t *sll, int (*cmp)(const void *, const void *))
{
	sll_node_t *q = sll->head, *t, *r, *s; 

	while(q->next) {
		for(r=t=q; r->next; r = r->next) 
			if( (*cmp)(t->data, r->next->data) >= 0 ) {
				t = r->next, s = r;
			}
		r->next = t->next;
		s = sll->head;
		sll->head = t;	
		if(t == q) {
			s->next = t->next, t->next = q->next, q->next = t;
		}
		if(t != sll->head) 
			t->next = s;
	}
}
static sll_t *merge2link(sll_t *p, sll_t *q, int (*cmp)(const void *, const void *))
{
	sll_node_t *t, *a = p->head, *b = q->head;
	sll_t *h;

	if(!b)
		return p;
	if(!a)
		return q;
	if( (*cmp)(&a->data, &b->data) < 0 )
		h = p, t = a, a = a->next;
	else
		h = q, t = b, b = b->next;
	while(a || b) 
		if( (*cmp)(&a->data, &b->data) < 0 )
			t = t->next = a, a = a->next;
		else
			t = t->next = b, b = b->next;
	if(a) {
		t->next = a;
		h->tail = p->tail;
	} else {
		t->next = b;
		h->tail = q->tail;
	}
	return h;
}

/* Merge asc/desc sorted links (all in same order) into a single sorted dst, 
 * various sll_t * parameters and a terminating NULL is expected. 
 * original ones will be destroyed automatically. */
void sll_merge(sll_t *dst, int (*cmp)(const void *, const void *), ...)
{

}

/* Traverse SLL via visit() from head to tail. */
void sll_trav(sll_t *dll, void (*visit)(const void *))
{

}

/* Free SLL allocated memory. */
void sll_destroy(sll_t *sll)
{
	sll_node_t *t;
	while(sll->head) {
		t = sll->head->next;
		free(sll->head);
		sll->head = t;
	}
	sll->tail = NULL, sll->linklen = 0;
}

#ifdef ADT_SLLIST_DEBUG
static void disp(sll_t *sll)
{
	int i;
	sll_node_t *t = sll->head;
	for(i=0; i<sll->linklen; ++i) {
		printf("data=%d\n", t->data.data);
		t = t->next;
	}
}
static bool test(const void *p)
{
	return sll->data > 100 ? true : false;
}
static bool test2(const void *p)
{
	return sll->data > 2 ? true : false;
}
static bool test3(const void *p)
{
	return sll->data==1 ? true : false;
}
static bool test4(const void *p)
{
	return sll->data==22 ? true : false;
}
int main(void)
{
	sll_t L,M, *p = &L, *dst = &M;
	sll_elem_t E, data = &E;
	sll_node_t *t;
	int i;
	
	sll_init(p);
	puts("adding tails...");
	for(i=0; i<10; ++i) {
		q->data = i;
		sll_addtail(p, q);
	}
	disp(p);

	sll_destroy(p);
	
	puts("\nadding heads...");
	for(i=0; i<5; ++i) {
		q->data = i;
		sll_addhead(p, q);
	}
	disp(p);
	
	puts("\nduplicating SLL...");
	sll_dup(dst, p);
	disp(dst);

	puts("\nreversing dst just duplicated...");
	sll_rev(dst);
	disp(dst);
	
	puts("\ndeleting head...");
	sll_delhead(dst);
	disp(dst);

	puts("\ndeleting tail...");
	sll_deltail(dst);
	disp(dst);	
	
	puts("\nadding tail...");
	q->data = 99;
	sll_addtail(dst, q);
	disp(dst);
	
	puts("\nadding head...");
	q->data = 14;
	sll_addhead(dst, q);
	disp(dst);	

	puts("\ndeleting matching nodes...");
	sll_node_cdel(dst, test);
	disp(dst);
	
	puts("\ndeleting matching nodes...");
	sll_node_cdel(dst, test2);
	disp(dst);

	puts("\nupdating matching nodes...");
	q->data = 22;
	sll_node_cupdt(p, test2, q);
	disp(p);
	
	puts("\ninserting before matching nodes...");
	q->data = 33;
	sll_node_bins(p, test3, q);
	disp(p);
	
	puts("\ninserting after matching nodes...");
	q->data = 44;
	sll_node_ains(p, test3, q);
	disp(p);
	
	sll_destroy(p);
	sll_destroy(dst);
	return 0;
}
#endif
